#!/usr/bin/env python
"""
Modelica parse Tree to AST tree.
"""
from __future__ import print_function, absolute_import, division, unicode_literals

import antlr4
import antlr4.Parser
from typing import Dict
from collections import deque, OrderedDict
import copy

from . import ast
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from .generated.ModelicaLexer import ModelicaLexer
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from .generated.ModelicaListener import ModelicaListener
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from .generated.ModelicaParser import ModelicaParser


# TODO
#  - Named function arguments (note that either all have to be named, or none)
#  - Make sure slice indices (eventually) evaluate to integers

class ModelicaFile:
    def __init__(self, **kwargs):
        self.within = []  # type: List[ComponentRef]
        self.classes = OrderedDict()  # type: OrderedDict[str, Class]
        super().__init__(**kwargs)


# noinspection PyPep8Naming
class ASTListener(ModelicaListener):
    def __init__(self):
        self.file_node = None  # type: ModelicaFile
        self.ast = {}  # type: Dict[ast.Node]
        self.ast_result = None  # type: ast.Node
        self.class_nodes = deque([ast.Class()])  # type: deque[ast.Class]
        self.comp_clause = None  # type: ast.ComponentClause
        self.eq_sect = None  # type: ast.EquationSection
        self.alg_sect = None  # type: ast.AlgorithmSection
        self.symbol_node = None  # type: ast.Symbol
        self.eq_comment = None  # type: str
        self.sym_count = 0  # type: int
        self.in_extends_clause = False

    @property
    def class_node(self):
        return self.class_nodes[-1]

    # FILE ===========================================================

    def enterStored_definition(self, ctx: ModelicaParser.Stored_definitionContext):

        file_node = ModelicaFile()
        self.ast[ctx] = file_node
        self.file_node = file_node

    def exitStored_definition(self, ctx: ModelicaParser.Stored_definitionContext):
        within = []
        if ctx.component_reference() is not None:
            within = [self.ast[ctx.component_reference()]]
        self.file_node.within = within

        for class_node in [self.ast[e] for e in ctx.stored_definition_class()]:
            self.ast[ctx].classes[class_node.name] = class_node
        self.ast_result = self.ast[ctx]

    # CLASS ===========================================================

    def exitStored_definition_class(self, ctx: ModelicaParser.Stored_definition_classContext):
        class_node = self.ast[ctx.class_definition()]
        class_node.final = ctx.FINAL() is not None
        self.ast[ctx] = class_node

    def enterClass_definition(self, ctx: ModelicaParser.Class_definitionContext):
        class_node = ast.Class()
        class_node.encapsulated = ctx.ENCAPSULATED() is not None
        class_node.partial = ctx.class_prefixes().PARTIAL() is not None
        class_node.type = ctx.class_prefixes().class_type().getText()

        self.class_nodes.append(class_node)

        self.ast[ctx] = class_node

    def exitClass_definition(self, ctx: ModelicaParser.Class_definitionContext):
        class_node = self.class_nodes.pop()
        self.class_node.classes[class_node.name] = class_node

    def exitShort_class_definition(self, ctx):
        self.ast[ctx] = ast.ShortClassDefinition(name=ctx.IDENT().getText(),
                                                 type=ctx.class_prefixes().class_type().getText(),
                                                 component=self.ast[ctx.component_reference()])

    def exitClass_spec_comp(self, ctx: ModelicaParser.Class_spec_compContext):
        class_node = self.class_node
        class_node.name = ctx.IDENT()[0].getText()
        class_node.comment = self.ast[ctx.string_comment()]

    def exitClass_spec_base(self, ctx: ModelicaParser.Class_spec_baseContext):
        class_node = self.class_node
        class_node.name = ctx.IDENT().getText()
        class_node.comment = self.ast[ctx.comment()]

        if ctx.class_modification() is not None:
            class_modification = self.ast[ctx.class_modification()]
        else:
            class_modification = ast.ClassModification()
        extends_clause = ast.ExtendsClause(component=self.ast[ctx.component_reference()],
                                          class_modification=class_modification)
        class_node.extends.append(extends_clause)

    def exitComposition(self, ctx: ModelicaParser.CompositionContext):
        for clause in self.ast[ctx.epriv]:
            if isinstance(clause, ast.ComponentClause):
                for symbol in clause.symbol_list:
                    symbol.visibility = ast.Visibility.PRIVATE
            elif isinstance(clause, ast.ExtendsClause):
                clause.visibility = ast.Visibility.PRIVATE

        if ctx.epub is not None:
            for clause in self.ast[ctx.epub]:
                if isinstance(clause, ast.ComponentClause):
                    for symbol in clause.symbol_list:
                        symbol.visibility = ast.Visibility.PUBLIC
                elif isinstance(clause, ast.ExtendsClause):
                    clause.visibility = ast.Visibility.PUBLIC

        if ctx.epro is not None:
            for clause in self.ast[ctx.epro]:
                if isinstance(clause, ast.ComponentClause):
                    for symbol in clause.symbol_list:
                        symbol.visibility = ast.Visibility.PROTECTED
                elif isinstance(clause, ast.ExtendsClause):
                    clause.visibility = ast.Visibility.PROTECTED

        for eqlist in [self.ast[e] for e in ctx.equation_section()]:
            if eqlist is not None:
                if eqlist.initial:
                    self.class_node.initial_equations += eqlist.equations
                else:
                    self.class_node.equations += eqlist.equations

        for alglist in [self.ast[e] for e in ctx.algorithm_section()]:
            if alglist is not None:
                if alglist.initial:
                    self.class_node.initial_statements += alglist.statements
                else:
                    self.class_node.statements += alglist.statements

        if ctx.comp_annotation is not None:
            self.class_node.annotation = self.ast[ctx.comp_annotation]

    def exitArgument(self, ctx: ModelicaParser.ArgumentContext):
        argument = ast.ClassModificationArgument()
        if ctx.element_modification_or_replaceable() is not None:
            argument.value = self.ast[ctx.element_modification_or_replaceable()]
            argument.redeclare = False
        else:
            argument.value = self.ast[ctx.element_redeclaration()]
            argument.redeclare = True
        self.ast[ctx] = argument

    def exitArgument_list(self, ctx: ModelicaParser.Argument_listContext):
        self.ast[ctx] = [self.ast[a] for a in ctx.argument()]

    def exitClass_modification(self, ctx: ModelicaParser.Class_modificationContext):
        arguments = []
        if ctx.argument_list() is not None:
            arguments = self.ast[ctx.argument_list()]
        self.ast[ctx] = ast.ClassModification(arguments=arguments)

    def enterEquation_section(self, ctx: ModelicaParser.Equation_sectionContext):
        eq_sect = ast.EquationSection(
            initial=ctx.INITIAL() is not None
        )
        self.ast[ctx] = eq_sect
        self.eq_sect = eq_sect

    def exitEquation_section(self, ctx: ModelicaParser.Equation_sectionContext):
        eq_sect = self.ast[ctx]
        if eq_sect.initial:
            eq_sect.equations.extend(self.ast[ctx.equation_block()])
        else:
            eq_sect.equations.extend(self.ast[ctx.equation_block()])

    def exitEquation_block(self, ctx: ModelicaParser.Equation_blockContext):
        self.ast[ctx] = [self.ast[e] for e in ctx.equation()]

    def exitStatement_block(self, ctx):
        self.ast[ctx] = [self.ast[e] for e in ctx.statement()]

    def enterAlgorithm_section(self, ctx: ModelicaParser.Algorithm_sectionContext):
        alg_sect = ast.AlgorithmSection(
            initial=ctx.INITIAL() is not None
        )
        self.ast[ctx] = alg_sect
        self.alg_sect = alg_sect

    def exitAlgorithm_section(self, ctx: ModelicaParser.Algorithm_sectionContext):
        alg_sect = self.ast[ctx]
        if alg_sect.initial:
            alg_sect.statements.extend(self.ast[ctx.statement_block()])
        else:
            alg_sect.statements.extend(self.ast[ctx.statement_block()])

    # EQUATION ===========================================================

    def enterEquation(self, ctx: ModelicaParser.EquationContext):
        pass

    def exitEquation(self, ctx):
        self.ast[ctx] = self.ast[ctx.equation_options()]
        try:
            self.ast[ctx].comment = self.ast[ctx.comment()]
        except AttributeError:
            pass

    def exitEquation_simple(self, ctx: ModelicaParser.Equation_simpleContext):
        self.ast[ctx] = ast.Equation(
            left=self.ast[ctx.simple_expression()],
            right=self.ast[ctx.expression()])

    def exitEquation_if(self, ctx: ModelicaParser.Equation_ifContext):
        self.ast[ctx] = self.ast[ctx.if_equation()]

    def exitEquation_for(self, ctx: ModelicaParser.Equation_forContext):
        self.ast[ctx] = self.ast[ctx.for_equation()]

    def exitEquation_connect_clause(self, ctx: ModelicaParser.Equation_connect_clauseContext):
        self.ast[ctx] = self.ast[ctx.connect_clause()]

    def exitArgument_expression(self, ctx: ModelicaParser.Argument_expressionContext):
        self.ast[ctx] = self.ast[ctx.expression()]

    def exitIf_equation(self, ctx: ModelicaParser.If_equationContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.IfEquation(
            conditions=conditions,
            blocks=blocks)

    def exitWhen_equation(self, ctx: ModelicaParser.When_equationContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.WhenEquation(
            conditions=conditions,
            blocks=blocks)

    def exitFor_equation(self, ctx: ModelicaParser.For_equationContext):
        self.ast[ctx] = ast.ForEquation(
            indices=self.ast[ctx.for_indices()],
            equations =self.ast[ctx.equation_block()])

    def exitConnect_clause(self, ctx: ModelicaParser.Connect_clauseContext):
        self.ast[ctx] = ast.ConnectClause(
            left=self.ast[ctx.component_reference()[0]],
            right=self.ast[ctx.component_reference()[1]])

    # STATEMENT ==========================================================

    # TODO:
    # - Missing statement types:
    #   + Function calls (inside current function).
    #   + Break
    #   + Return
    #   + While
    #   + When? (also in equation missing)

    def enterStatement(self, ctx: ModelicaParser.StatementContext):
        pass

    def exitStatement(self, ctx: ModelicaParser.StatementContext):
        self.ast[ctx] = self.ast[ctx.statement_options()]
        try:
            self.ast[ctx].comment = self.ast[ctx.comment()]
        except AttributeError:
            pass

    def exitStatement_component_reference(self, ctx: ModelicaParser.Statement_component_referenceContext):
        self.ast[ctx] = ast.AssignmentStatement(
            left=[self.ast[ctx.component_reference()]],
            right=self.ast[ctx.expression()])

    def exitStatement_component_function(self, ctx: ModelicaParser.Statement_component_functionContext):
        all_comp_refs = [self.ast[x] for x in ctx.component_reference()]

        right = ast.Expression(
            operator=all_comp_refs[-1],
            operands=[self.ast[x.expression()]
                      for x in ctx.function_call_args().function_arguments().function_argument()]
        )

        self.ast[ctx] = ast.AssignmentStatement(
            left=all_comp_refs[:-1],
            right=right)

    def exitStatement_if(self, ctx: ModelicaParser.Statement_ifContext):
        self.ast[ctx] = self.ast[ctx.if_statement()]

    def exitStatement_for(self, ctx: ModelicaParser.Statement_forContext):
        self.ast[ctx] = self.ast[ctx.for_statement()]

    def exitStatement_when(self, ctx: ModelicaParser.Equation_whenContext):
        self.ast[ctx] = self.ast[ctx.when_equation()]

    def exitIf_statement(self, ctx: ModelicaParser.If_statementContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.IfStatement(
            conditions=conditions,
            blocks=blocks)

    def exitWhen_statement(self, ctx: ModelicaParser.When_statementContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.WhenStatement(
            conditions=conditions,
            blocks=blocks)

    def exitFor_statement(self, ctx: ModelicaParser.For_statementContext):
        self.ast[ctx] = ast.ForStatement(
            indices=self.ast[ctx.for_indices()],
            statements=self.ast[ctx.statement_block()])

    # EXPRESSIONS ===========================================================

    def exitSimple_expression(self, ctx: ModelicaParser.Simple_expressionContext):
        if len(ctx.expr()) > 1:
            if len(ctx.expr()) > 2:
                step = self.ast[ctx.expr()[2]]
            else:
                step = ast.Primary(value=1)
            self.ast[ctx] = ast.Slice(start=self.ast[ctx.expr()[0]], stop=self.ast[ctx.expr()[1]], step=step)
        else:
            self.ast[ctx] = self.ast[ctx.expr()[0]]

    def exitExpression_simple(self, ctx: ModelicaParser.Expression_simpleContext):
        self.ast[ctx] = self.ast[ctx.simple_expression()]

    def exitExpression_if(self, ctx: ModelicaParser.Expression_ifContext):
        all_expr = [self.ast[s] for s in ctx.expression()]
        # Note that an else block is guaranteed to exist.
        conditions = all_expr[:-1:2]
        expressions = all_expr[1::2] + all_expr[-1:]

        self.ast[ctx] = ast.IfExpression(
            conditions=conditions,
            expressions=expressions)

    def exitExpr_primary(self, ctx: ModelicaParser.Expr_primaryContext):
        self.ast[ctx] = self.ast[ctx.primary()]

    def exitExpr_add(self, ctx: ModelicaParser.Expr_addContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_exp(self, ctx: ModelicaParser.Expr_expContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.primary()]
        )

    def exitExpr_mul(self, ctx: ModelicaParser.Expr_mulContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_rel(self, ctx: ModelicaParser.Expr_relContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_not(self, ctx: ModelicaParser.Expr_notContext):
        self.ast[ctx] = ast.Expression(
            operator='not',
            operands=[self.ast[ctx.expr()]]
        )

    def exitExpr_and(self, ctx: ModelicaParser.Expr_andContext):
        self.ast[ctx] = ast.Expression(
            operator='and',
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_or(self, ctx: ModelicaParser.Expr_orContext):
        self.ast[ctx] = ast.Expression(
            operator='or',
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_neg(self, ctx: ModelicaParser.Expr_negContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[ctx.expr()]]
        )

    def exitSubscript(self, ctx: ModelicaParser.SubscriptContext):
        if ctx.expression() is not None:
            self.ast[ctx] = self.ast[ctx.expression()]
        else:
            self.ast[ctx] = ast.Slice()

    def exitArray_subscripts(self, ctx: ModelicaParser.Array_subscriptsContext):
        self.ast[ctx] = [self.ast[s] for s in ctx.subscript()]

    def exitFor_index(self, ctx):
        self.ast[ctx] = ast.ForIndex(name=ctx.IDENT().getText(), expression=self.ast[ctx.expression()])

    def exitFor_indices(self, ctx: ModelicaParser.For_indicesContext):
        self.ast[ctx] = [self.ast[s] for s in ctx.for_index()]

    # PRIMARY ===========================================================

    def exitPrimary_unsigned_number(self, ctx: ModelicaParser.Primary_unsigned_numberContext):
        self.ast[ctx] = ast.Primary(value=float(ctx.getText()))

    def exitPrimary_string(self, ctx: ModelicaParser.Primary_stringContext):
        val = ctx.getText()
        assert val.startswith('"') and val.endswith('"')
        self.ast[ctx] = ast.Primary(value=val[1:-1])

    def exitPrimary_false(self, ctx: ModelicaParser.Primary_falseContext):
        self.ast[ctx] = ast.Primary(value=False)

    def exitPrimary_true(self, ctx: ModelicaParser.Primary_trueContext):
        self.ast[ctx] = ast.Primary(value=True)

    def exitPrimary_function(self, ctx: ModelicaParser.Primary_functionContext):
        # TODO: Could possible be cleaner if we let the expression in the ast bubble up.
        #       E.g. self.ast[x] below, instead of self.ast[x.expression].
        self.ast[ctx] = ast.Expression(
            operator=self.ast[ctx.component_reference()],
            operands=[self.ast[x.expression()]
                      for x in ctx.function_call_args().function_arguments().function_argument()]
        )

    def exitPrimary_derivative(self, ctx: ModelicaParser.Primary_derivativeContext):
        self.ast[ctx] = ast.Expression(
            operator='der',
            operands=[self.ast[x.expression()]
                      for x in ctx.function_call_args().function_arguments().function_argument()]
        )
        # TODO 'state' is not a standard prefix;  disable this for now as it does not work
        # when differentiating states defined in superclasses.
        # if 'state' not in self.class_node.symbols[comp_name].prefixes:
        #    self.class_node.symbols[comp_name].prefixes += ['state']

    def exitType_specifier_element(self, ctx: ModelicaParser.Type_specifier_elementContext):
        self.ast[ctx] = ast.ComponentRef(
            name=ctx.IDENT().getText(),
            indices=[[None]],
            child=[]
        )

    def exitType_specifier(self, ctx: ModelicaParser.Type_specifierContext):
        for element in reversed([self.ast[ctx] for ctx in ctx.type_specifier_element()]):
            if ctx in self.ast:
                element.child = [self.ast[ctx]]
            self.ast[ctx] = element

    def exitComponent_reference_element(self, ctx: ModelicaParser.Component_reference_elementContext):
        if ctx.array_subscripts() is not None:
            indices = [[self.ast[x] for x in ctx.array_subscripts().subscript()]]
        else:
            indices = [[None]]
        self.ast[ctx] = ast.ComponentRef(
            name=ctx.IDENT().getText(),
            indices=indices,
            child=[]
        )

    def exitComponent_reference(self, ctx: ModelicaParser.Component_referenceContext):
        for element in reversed([self.ast[ctx] for ctx in ctx.component_reference_element()]):
            if ctx in self.ast:
                element.child = [self.ast[ctx]]
            self.ast[ctx] = element

    def exitPrimary_component_reference(self, ctx: ModelicaParser.Primary_component_referenceContext):
        self.ast[ctx] = self.ast[ctx.component_reference()]

    def exitPrimary_output_expression_list(self, ctx: ModelicaParser.Primary_output_expression_listContext):
        self.ast[ctx] = [self.ast[x] for x in ctx.output_expression_list().expression()]
        # Collapse lists containing a single expression
        if len(self.ast[ctx]) == 1:
            self.ast[ctx] = self.ast[ctx][0]

    def exitPrimary_function_arguments(self, ctx: ModelicaParser.Primary_function_argumentsContext):
        # TODO: This does not support for generators yet.
        #       Only expressions are supported, e.g. {1.0, 2.0, 3.0}.
        v = [self.ast[x.expression()] for x in ctx.function_arguments().function_argument()]
        self.ast[ctx] = ast.Array(values=v)

    def exitEquation_function(self, ctx: ModelicaParser.Equation_functionContext):
        self.ast[ctx] = ast.Function(
            name=ctx.name().getText(),
            arguments=[self.ast[x.expression()]
                for x in ctx.function_call_args().function_arguments().function_argument()]
        )

    def exitEquation_when(self, ctx: ModelicaParser.Equation_whenContext):
        self.ast[ctx] = self.ast[ctx.when_equation()]

    # COMPONENTS ===========================================================

    def exitElement_list(self, ctx: ModelicaParser.Element_listContext):
        self.ast[ctx] = [self.ast[e] for e in ctx.element()]

    def exitElement(self, ctx: ModelicaParser.ElementContext):
        self.ast[ctx] = self.ast[ctx.getChild(ctx.getAltNumber())]

    def exitImport_list(self, ctx: ModelicaParser.Import_listContext):
        self.ast[ctx] = [ctx.IDENT()] + self.ast[ctx.import_list()]

    def exitImport_clause(self, ctx: ModelicaParser.Import_clauseContext):
        component = self.ast[ctx.component_reference()]
        if ctx.IDENT() is not None:
            self.ast[ctx] = ast.ImportAsClause(component=component, name=ctx.IDENT().getText())
        else:
            symbols = self.ast[ctx.import_list()]
            self.ast[ctx] = ast.ImportFromClause(component=component, symbols=symbols)
        self.class_node.imports += [self.ast[ctx]]

    def enterExtends_clause(self, ctx: ModelicaParser.Extends_clauseContext):
        self.in_extends_clause = True

    def exitExtends_clause(self, ctx: ModelicaParser.Extends_clauseContext):
        if ctx.class_modification() is not None:
            class_modification = self.ast[ctx.class_modification()]
        else:
            class_modification = ast.ClassModification()
        self.ast[ctx] = ast.ExtendsClause(component=self.ast[ctx.component_reference()],
                                          class_modification=class_modification)
        self.class_node.extends += [self.ast[ctx]]

        self.in_extends_clause = False

    def exitRegular_element(self, ctx: ModelicaParser.Regular_elementContext):
        if ctx.comp_elem is not None:
            self.ast[ctx] = self.ast[ctx.comp_elem]
        else:
            self.ast[ctx] = self.ast[ctx.class_elem]

    def exitReplaceable_element(self, ctx: ModelicaParser.Replaceable_elementContext):
        if ctx.comp_elem is not None:
            self.ast[ctx] = self.ast[ctx.comp_elem]
        else:
            self.ast[ctx] = self.ast[ctx.class_elem]

    def enterComponent_clause(self, ctx: ModelicaParser.Component_clauseContext):
        prefixes = ctx.type_prefix().getText().split(' ')
        if prefixes[0] == '':
            prefixes = []
        self.ast[ctx] = ast.ComponentClause(
            prefixes=prefixes,
        )
        self.comp_clause = self.ast[ctx]

    def enterComponent_clause1(self, ctx: ModelicaParser.Component_clause1Context):
        prefixes = ctx.type_prefix().getText().split(' ')
        if prefixes[0] == '':
            prefixes = []
        self.ast[ctx] = ast.ComponentClause(
            prefixes=prefixes,
        )
        self.comp_clause = self.ast[ctx]

    def exitComponent_clause(self, ctx: ModelicaParser.Component_clauseContext):
        clause = self.ast[ctx]
        # The component clause and all its symbols share the same type.
        # However, the type will only be turned into a component reference
        # somewhere between the enterDeclaration and exitDeclaration functions
        # of the symbols. Therefore, we need to keep the component clause's
        # type, and all its symbols' types, pointing at the same empty
        # (ComponentRef) object until we can fill it.
        clause.type.__dict__.update(self.ast[ctx.type_specifier()].__dict__)
        if ctx.array_subscripts() is not None:
            clause.dimensions = [self.ast[ctx.array_subscripts()]]
            for sym in self.comp_clause.symbol_list:
                s = self.class_node.symbols[sym.name]
                s.dimensions = clause.dimensions

        # We make sure that all references to the objects are unique per
        # symbol making copies. Note that if there is only one symbol in the
        # component clause, it is already unique.
        for sym in self.comp_clause.symbol_list[1:]:
            s = self.class_node.symbols[sym.name]
            s.dimensions = list(s.dimensions)
            s.prefixes = list(s.prefixes)
            s.type = copy.deepcopy(clause.type)

    def exitComponent_clause1(self, ctx: ModelicaParser.Component_clause1Context):
        clause = self.ast[ctx]
        clause.type.__dict__.update(self.ast[ctx.type_specifier()].__dict__)

        for sym in self.comp_clause.symbol_list[1:]:
            s = self.class_node.symbols[sym.name]
            s.dimensions = list(s.dimensions)
            s.prefixes = list(s.prefixes)
            s.type = copy.deepcopy(clause.type)

    def enterComponent_declaration(self, ctx: ModelicaParser.Component_declarationContext):
        sym = ast.Symbol(order=self.sym_count)
        self.sym_count += 1
        self.ast[ctx] = sym
        self.symbol_node = sym
        self.comp_clause.symbol_list += [sym]

    def enterComponent_declaration1(self, ctx: ModelicaParser.Component_declaration1Context):
        sym = ast.Symbol(order=self.sym_count)
        self.sym_count += 1
        self.ast[ctx] = sym
        self.symbol_node = sym
        self.comp_clause.symbol_list += [sym]

    def enterElement_modification(self, ctx: ModelicaParser.Element_modificationContext):
        if self.symbol_node is not None:
            self.ast[ctx] = self.symbol_node
        else:
            sym = ast.Symbol(order=self.sym_count)
            self.sym_count += 1
            self.ast[ctx] = sym
            self.symbol_node = sym

    def exitComponent_declaration(self, ctx: ModelicaParser.Component_declarationContext):
        self.ast[ctx].comment = self.ast[ctx.comment()]
        self.symbol_node = None

    def exitComponent_declaration1(self, ctx: ModelicaParser.Component_declaration1Context):
        self.ast[ctx].comment = self.ast[ctx.comment()]
        self.symbol_node = None

    def enterDeclaration(self, ctx: ModelicaParser.DeclarationContext):
        sym = self.symbol_node
        dimensions = None
        if self.comp_clause.dimensions is not None:
            dimensions = self.comp_clause.dimensions
        sym.name = ctx.IDENT().getText()
        sym.dimensions = dimensions
        sym.prefixes = self.comp_clause.prefixes
        sym.type = self.comp_clause.type

        # Declarations can also occur in extends clauses, in which case we do not have to add it to the class's symbols.
        if not self.in_extends_clause:
            if sym.name in self.class_node.symbols:
                raise IOError(sym.name, 'already defined')
            self.class_node.symbols[sym.name] = sym

    def exitDeclaration(self, ctx: ModelicaParser.DeclarationContext):
        sym = self.symbol_node
        if ctx.array_subscripts() is not None:
            sym.dimensions = [self.ast[ctx.array_subscripts()]]
        if ctx.modification() is not None:
            for mod in self.ast[ctx.modification()]:
                if isinstance(mod, ast.ClassModification):
                    sym.class_modification = mod
                else:
                    # Assignment of value, which we turn into a modification here.
                    vmod_arg = ast.ClassModificationArgument()
                    vmod_arg.value = ast.ElementModification()
                    vmod_arg.value.component = ast.ComponentRef(name="value")
                    vmod_arg.value.modifications = [mod]

                    if sym.class_modification is None:
                        sym_mod = ast.ClassModification()
                        sym_mod.arguments.append(vmod_arg)
                        sym.class_modification = sym_mod
                    else:
                        sym.class_modification.arguments.append(vmod_arg)

    def exitElement_modification(self, ctx: ModelicaParser.Element_modificationContext):
        component = self.ast[ctx.component_reference()]
        if ctx.modification() is not None:
            modifications = self.ast[ctx.modification()]
        else:
            modifications = []

        self.ast[ctx] = ast.ElementModification(component=component, modifications=modifications)

    def exitModification_class(self, ctx: ModelicaParser.Modification_classContext):
        self.ast[ctx] = [self.ast[ctx.class_modification()]]
        if ctx.expression() is not None:
            self.ast[ctx] += [self.ast[ctx.expression()]]

    def exitModification_assignment(self, ctx: ModelicaParser.Modification_assignmentContext):
        self.ast[ctx] = [self.ast[ctx.expression()]]

    def exitModification_assignment2(self, ctx: ModelicaParser.Modification_assignment2Context):
        self.ast[ctx] = [self.ast[ctx.expression()]]

    def exitElement_replaceable(self, ctx: ModelicaParser.Element_replaceableContext):
        if ctx.component_clause1() is not None:
            self.ast[ctx] = self.ast[ctx.component_clause1()]
        else:
            self.ast[ctx] = self.ast[ctx.short_class_definition()]

    def exitElement_modification_or_replaceable(self, ctx: ModelicaParser.Element_modification_or_replaceableContext):
        if ctx.element_modification() is not None:
            self.ast[ctx] = self.ast[ctx.element_modification()]
        else:
            self.ast[ctx] = self.ast[ctx.element_replaceable()]

    def exitElement_redeclaration(self, ctx: ModelicaParser.Element_redeclarationContext):
        if ctx.component_clause1() is not None:
            self.ast[ctx] = self.ast[ctx.component_clause1()]
        else:
            self.ast[ctx] = self.ast[ctx.short_class_definition()]

    # COMMENTS ==============================================================

    def exitComment(self, ctx: ModelicaParser.CommentContext):
        self.ast[ctx] = self.ast[ctx.string_comment()]

    def exitString_comment(self, ctx: ModelicaParser.String_commentContext):
        self.ast[ctx] = ctx.getText()[1:-1]

    # ANNOTATIONS ==========================================================

    def exitAnnotation(self, ctx: ModelicaParser.AnnotationContext):
        self.ast[ctx] = self.ast[ctx.class_modification()]


# UTILITY FUNCTIONS ========================================================
def file_to_tree(f: ModelicaFile) -> ast.Tree:
    # TODO: We can only insert where classes exist. For example, if we have a
    # within statement, we have to check if the nodes of the within statement
    # are actually in the tree, and if not raise an exception.
    root = ast.Tree()
    insert_node = root
    if f.within:
        for p in f.within[0].to_tuple():
            package = ast.Class(name=p, type="package")
            insert_node.classes[p] = package
            insert_node = package

    insert_node.classes.update(f.classes)

    root.update_parent_refs()

    return root

def parse(text):
    input_stream = antlr4.InputStream(text)
    lexer = ModelicaLexer(input_stream)
    stream = antlr4.CommonTokenStream(lexer)
    parser = ModelicaParser(stream)
    # parser.buildParseTrees = False
    parse_tree = parser.stored_definition()
    ast_listener = ASTListener()
    parse_walker = antlr4.ParseTreeWalker()
    parse_walker.walk(ast_listener, parse_tree)
    modelica_file = ast_listener.ast_result
    return file_to_tree(modelica_file)



from __future__ import print_function, absolute_import, division, print_function, unicode_literals

import copy
import os
from typing import List

import jinja2

from pymoca import ast
from pymoca.tree import TreeListener, TreeWalker, flatten

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
# noinspection PyUnresolvedReferences
BUILTINS = dir(__builtins__) + ['psi']


class SympyGenerator(TreeListener):
    def __init__(self):
        super(SympyGenerator, self).__init__()
        self.src = {}

    def exitTree(self, tree: ast.Tree):
        d = {'classes': []}
        for key in sorted(tree.classes.keys()):
            d['classes'] += [self.src[tree.classes[key]]]

        template = jinja2.Template('''
# do not edit, generated by pymoca

from __future__ import print_function, division
import sympy
import sympy.physics.mechanics as mech
from pymoca.backends.sympy.runtime import OdeModel
from sympy import sin, cos, tan
{%- for class_key, class in tree.classes.items() %}
{{ render.src[class] }}
{%- endfor %}
''')
        self.src[tree] = template.render({
            'tree': tree,
            'render': self,
        })

    def exitClass(self, tree: ast.Class):
        states = []
        inputs = []
        outputs = []
        constants = []
        parameters = []
        variables = []

        symbols = sorted(tree.symbols.values(), key=lambda x: x.order)  # type: List[ast.Symbol]

        for s in symbols:

            if len(s.prefixes) == 0:
                variables += [s]
            else:
                for prefix in s.prefixes:
                    if prefix == 'state':
                        states += [s]
                    elif prefix == 'constant':
                        constants += [s]
                    elif prefix == 'parameter':
                        parameters += [s]
                    elif prefix == 'input':
                        inputs += [s]
                    elif prefix == 'output':
                        outputs += [s]

        for s in outputs:
            if s not in states:
                variables += [s]

        states_str = ', '.join([self.src[s] for s in states])
        inputs_str = ', '.join([self.src[s] for s in inputs])
        outputs_str = ', '.join([self.src[s] for s in outputs])
        constants_str = ', '.join([self.src[s] for s in constants])
        parameters_str = ', '.join([self.src[s] for s in parameters])
        variables_str = ', '.join([self.src[s] for s in variables])

        d = locals()
        d.pop('self')
        d['render'] = self

        template = jinja2.Template('''

class {{tree.name}}(OdeModel):

    def __init__(self):

        super({{tree.name}}, self).__init__()

        # states
        {% if states_str|length > 0 -%}
        {{ states_str }} = mech.dynamicsymbols('{{ states_str|replace('__', '.') }}')
        {% endif -%}
        self.x = sympy.Matrix([{{ states_str }}])
        self.x0 = {
            {% for s in states -%}
            {{render.src[s]}} : {{tree.symbols[s.name].value.value if tree.symbols[s.name].value.value else tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # variables
        {% if variables_str|length > 0 -%}
        {{ variables_str }} = mech.dynamicsymbols('{{ variables_str|replace('__', '.') }}')
        {% endif -%}
        self.v = sympy.Matrix([{{ variables_str }}])

        # constants
        {% if constants_str|length > 0 -%}
        {{ constants_str }} = sympy.symbols('{{ constants_str|replace('__', '.') }}')
        {% endif -%}
        self.c = sympy.Matrix([{{ constants_str }}])
        self.c0 = {
            {% for s in constants -%}
            {{render.src[s]}} : {{tree.symbols[s.name].value.value if tree.symbols[s.name].value.value else tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # parameters
        {% if parameters_str|length > 0 -%}
        {{ parameters_str }} = sympy.symbols('{{ parameters_str|replace('__', '.') }}')
        {% endif -%}
        self.p = sympy.Matrix([{{ parameters_str }}])
        self.p0 = {
            {% for s in parameters -%}
            {{render.src[s]}} : {{tree.symbols[s.name].value.value if tree.symbols[s.name].value.value else tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # inputs
        {% if inputs_str|length > 0 -%}
        {{ inputs_str }} = mech.dynamicsymbols('{{ inputs_str|replace('__', '.') }}')
        {% endif -%}
        self.u = sympy.Matrix([{{ inputs_str }}])
        self.u0 = {
            {% for s in inputs -%}
            {{render.src[s]}} : {{tree.symbols[s.name].value.value if tree.symbols[s.name].value.value else tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # outputs
        {% if outputs_str|length > 0 -%}
        {{ outputs_str }} = mech.dynamicsymbols('{{ outputs_str|replace('__', '.') }}')
        {% endif -%}
        self.y = sympy.Matrix([{{ outputs_str }}])

        # equations
        self.eqs = [
            {% for eq in tree.equations -%}
            {{ render.src[eq] }},
            {% endfor -%}
        ]

        self.compute_fg()
''')
        self.src[tree] = template.render(d)

    def exitExpression(self, tree: ast.Expression):
        op = str(tree.operator)
        n_operands = len(tree.operands)
        if op == 'der':
            src = '({var:s}).diff(self.t)'.format(
                var=self.src[tree.operands[0]])
        elif op in ['*', '+', '-', '/'] and n_operands == 2:
            src = '{left:s} {op:s} {right:s}'.format(
                op=op,
                left=self.src[tree.operands[0]],
                right=self.src[tree.operands[1]])
        elif op in ['+', '-'] and n_operands == 1:
            src = '{op:s} {expr:s}'.format(
                op=op,
                expr=self.src[tree.operands[0]])
        else:
            operand_src = ','.join([self.src[o] for o in tree.operands])
            src = "{tree.operator.name:s}({operand_src:s})".format(**locals())
        self.src[tree] = src

    def exitPrimary(self, tree: ast.Primary):
        val = str(tree.value)
        self.src[tree] = "{:s}".format(val)

    def exitComponentRef(self, tree: ast.ComponentRef):

        # prevent name clash with builtins
        name = tree.name.replace('.', '__')
        while name in BUILTINS:
            name = name + '_'
        self.src[tree] = name

    def exitSymbol(self, tree: ast.Symbol):
        # prevent name clash with builtins
        name = tree.name.replace('.', '__')
        while name in BUILTINS:
            name = name + '_'

        self.src[tree] = name

    def exitEquation(self, tree: ast.Equation):
        self.src[tree] = "{left:s} - ({right:s})".format(
            left=self.src[tree.left],
            right=self.src[tree.right])


def generate(ast_tree: ast.Tree, model_name: str):
    """
    :param ast_tree: AST to generate from
    :param model_name: class to generate
    :return: sympy source code for model
    """
    component_ref = ast.ComponentRef.from_string(model_name)
    ast_tree_new = copy.deepcopy(ast_tree)
    ast_walker = TreeWalker()
    flat_tree = flatten(ast_tree_new, component_ref)
    sympy_gen = SympyGenerator()
    ast_walker.walk(sympy_gen, flat_tree)
    return sympy_gen.src[flat_tree]



# Generated from Modelica.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ModelicaParser import ModelicaParser
else:
    from ModelicaParser import ModelicaParser

# This class defines a complete listener for a parse tree produced by ModelicaParser.
class ModelicaListener(ParseTreeListener):

    # Enter a parse tree produced by ModelicaParser#stored_definition.
    def enterStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#stored_definition.
    def exitStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#stored_definition_class.
    def enterStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):
        pass

    # Exit a parse tree produced by ModelicaParser#stored_definition_class.
    def exitStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_definition.
    def enterClass_definition(self, ctx:ModelicaParser.Class_definitionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_definition.
    def exitClass_definition(self, ctx:ModelicaParser.Class_definitionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_prefixes.
    def enterClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_prefixes.
    def exitClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_type.
    def enterClass_type(self, ctx:ModelicaParser.Class_typeContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_type.
    def exitClass_type(self, ctx:ModelicaParser.Class_typeContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_spec_comp.
    def enterClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_spec_comp.
    def exitClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_spec_base.
    def enterClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_spec_base.
    def exitClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_spec_enum.
    def enterClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_spec_enum.
    def exitClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_spec_der.
    def enterClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_spec_der.
    def exitClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):
        pass


    # Enter a parse tree produced by ModelicaParser#class_spec_extends.
    def enterClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_spec_extends.
    def exitClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#base_prefix.
    def enterBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):
        pass

    # Exit a parse tree produced by ModelicaParser#base_prefix.
    def exitBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):
        pass


    # Enter a parse tree produced by ModelicaParser#enum_list.
    def enterEnum_list(self, ctx:ModelicaParser.Enum_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#enum_list.
    def exitEnum_list(self, ctx:ModelicaParser.Enum_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#enumeration_literal.
    def enterEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):
        pass

    # Exit a parse tree produced by ModelicaParser#enumeration_literal.
    def exitEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):
        pass


    # Enter a parse tree produced by ModelicaParser#composition.
    def enterComposition(self, ctx:ModelicaParser.CompositionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#composition.
    def exitComposition(self, ctx:ModelicaParser.CompositionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#language_specification.
    def enterLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#language_specification.
    def exitLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#external_function_call.
    def enterExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):
        pass

    # Exit a parse tree produced by ModelicaParser#external_function_call.
    def exitExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element_list.
    def enterElement_list(self, ctx:ModelicaParser.Element_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element_list.
    def exitElement_list(self, ctx:ModelicaParser.Element_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element.
    def enterElement(self, ctx:ModelicaParser.ElementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element.
    def exitElement(self, ctx:ModelicaParser.ElementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#regular_element.
    def enterRegular_element(self, ctx:ModelicaParser.Regular_elementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#regular_element.
    def exitRegular_element(self, ctx:ModelicaParser.Regular_elementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#replaceable_element.
    def enterReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#replaceable_element.
    def exitReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#import_clause.
    def enterImport_clause(self, ctx:ModelicaParser.Import_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#import_clause.
    def exitImport_clause(self, ctx:ModelicaParser.Import_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#import_list.
    def enterImport_list(self, ctx:ModelicaParser.Import_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#import_list.
    def exitImport_list(self, ctx:ModelicaParser.Import_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#extends_clause.
    def enterExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#extends_clause.
    def exitExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#constraining_clause.
    def enterConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#constraining_clause.
    def exitConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_clause.
    def enterComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#component_clause.
    def exitComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#type_prefix.
    def enterType_prefix(self, ctx:ModelicaParser.Type_prefixContext):
        pass

    # Exit a parse tree produced by ModelicaParser#type_prefix.
    def exitType_prefix(self, ctx:ModelicaParser.Type_prefixContext):
        pass


    # Enter a parse tree produced by ModelicaParser#type_specifier_element.
    def enterType_specifier_element(self, ctx:ModelicaParser.Type_specifier_elementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#type_specifier_element.
    def exitType_specifier_element(self, ctx:ModelicaParser.Type_specifier_elementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#type_specifier.
    def enterType_specifier(self, ctx:ModelicaParser.Type_specifierContext):
        pass

    # Exit a parse tree produced by ModelicaParser#type_specifier.
    def exitType_specifier(self, ctx:ModelicaParser.Type_specifierContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_list.
    def enterComponent_list(self, ctx:ModelicaParser.Component_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#component_list.
    def exitComponent_list(self, ctx:ModelicaParser.Component_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_declaration.
    def enterComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#component_declaration.
    def exitComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#condition_attribute.
    def enterCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):
        pass

    # Exit a parse tree produced by ModelicaParser#condition_attribute.
    def exitCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):
        pass


    # Enter a parse tree produced by ModelicaParser#declaration.
    def enterDeclaration(self, ctx:ModelicaParser.DeclarationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#declaration.
    def exitDeclaration(self, ctx:ModelicaParser.DeclarationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#modification_class.
    def enterModification_class(self, ctx:ModelicaParser.Modification_classContext):
        pass

    # Exit a parse tree produced by ModelicaParser#modification_class.
    def exitModification_class(self, ctx:ModelicaParser.Modification_classContext):
        pass


    # Enter a parse tree produced by ModelicaParser#modification_assignment.
    def enterModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):
        pass

    # Exit a parse tree produced by ModelicaParser#modification_assignment.
    def exitModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):
        pass


    # Enter a parse tree produced by ModelicaParser#modification_assignment2.
    def enterModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):
        pass

    # Exit a parse tree produced by ModelicaParser#modification_assignment2.
    def exitModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):
        pass


    # Enter a parse tree produced by ModelicaParser#class_modification.
    def enterClass_modification(self, ctx:ModelicaParser.Class_modificationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#class_modification.
    def exitClass_modification(self, ctx:ModelicaParser.Class_modificationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#argument_list.
    def enterArgument_list(self, ctx:ModelicaParser.Argument_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#argument_list.
    def exitArgument_list(self, ctx:ModelicaParser.Argument_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#argument.
    def enterArgument(self, ctx:ModelicaParser.ArgumentContext):
        pass

    # Exit a parse tree produced by ModelicaParser#argument.
    def exitArgument(self, ctx:ModelicaParser.ArgumentContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element_modification_or_replaceable.
    def enterElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element_modification_or_replaceable.
    def exitElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element_modification.
    def enterElement_modification(self, ctx:ModelicaParser.Element_modificationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element_modification.
    def exitElement_modification(self, ctx:ModelicaParser.Element_modificationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element_redeclaration.
    def enterElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element_redeclaration.
    def exitElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#element_replaceable.
    def enterElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):
        pass

    # Exit a parse tree produced by ModelicaParser#element_replaceable.
    def exitElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_clause1.
    def enterComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):
        pass

    # Exit a parse tree produced by ModelicaParser#component_clause1.
    def exitComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):
        pass


    # Enter a parse tree produced by ModelicaParser#component_declaration1.
    def enterComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):
        pass

    # Exit a parse tree produced by ModelicaParser#component_declaration1.
    def exitComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):
        pass


    # Enter a parse tree produced by ModelicaParser#short_class_definition.
    def enterShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#short_class_definition.
    def exitShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_block.
    def enterEquation_block(self, ctx:ModelicaParser.Equation_blockContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_block.
    def exitEquation_block(self, ctx:ModelicaParser.Equation_blockContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_section.
    def enterEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_section.
    def exitEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_block.
    def enterStatement_block(self, ctx:ModelicaParser.Statement_blockContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_block.
    def exitStatement_block(self, ctx:ModelicaParser.Statement_blockContext):
        pass


    # Enter a parse tree produced by ModelicaParser#algorithm_section.
    def enterAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#algorithm_section.
    def exitAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_simple.
    def enterEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_simple.
    def exitEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_if.
    def enterEquation_if(self, ctx:ModelicaParser.Equation_ifContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_if.
    def exitEquation_if(self, ctx:ModelicaParser.Equation_ifContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_for.
    def enterEquation_for(self, ctx:ModelicaParser.Equation_forContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_for.
    def exitEquation_for(self, ctx:ModelicaParser.Equation_forContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_connect_clause.
    def enterEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_connect_clause.
    def exitEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_when.
    def enterEquation_when(self, ctx:ModelicaParser.Equation_whenContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_when.
    def exitEquation_when(self, ctx:ModelicaParser.Equation_whenContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation_function.
    def enterEquation_function(self, ctx:ModelicaParser.Equation_functionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation_function.
    def exitEquation_function(self, ctx:ModelicaParser.Equation_functionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#equation.
    def enterEquation(self, ctx:ModelicaParser.EquationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#equation.
    def exitEquation(self, ctx:ModelicaParser.EquationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_component_reference.
    def enterStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_component_reference.
    def exitStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_component_function.
    def enterStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_component_function.
    def exitStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_break.
    def enterStatement_break(self, ctx:ModelicaParser.Statement_breakContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_break.
    def exitStatement_break(self, ctx:ModelicaParser.Statement_breakContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_return.
    def enterStatement_return(self, ctx:ModelicaParser.Statement_returnContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_return.
    def exitStatement_return(self, ctx:ModelicaParser.Statement_returnContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_if.
    def enterStatement_if(self, ctx:ModelicaParser.Statement_ifContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_if.
    def exitStatement_if(self, ctx:ModelicaParser.Statement_ifContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_for.
    def enterStatement_for(self, ctx:ModelicaParser.Statement_forContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_for.
    def exitStatement_for(self, ctx:ModelicaParser.Statement_forContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_while.
    def enterStatement_while(self, ctx:ModelicaParser.Statement_whileContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_while.
    def exitStatement_while(self, ctx:ModelicaParser.Statement_whileContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement_when.
    def enterStatement_when(self, ctx:ModelicaParser.Statement_whenContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement_when.
    def exitStatement_when(self, ctx:ModelicaParser.Statement_whenContext):
        pass


    # Enter a parse tree produced by ModelicaParser#statement.
    def enterStatement(self, ctx:ModelicaParser.StatementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#statement.
    def exitStatement(self, ctx:ModelicaParser.StatementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#if_equation.
    def enterIf_equation(self, ctx:ModelicaParser.If_equationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#if_equation.
    def exitIf_equation(self, ctx:ModelicaParser.If_equationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#if_statement.
    def enterIf_statement(self, ctx:ModelicaParser.If_statementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#if_statement.
    def exitIf_statement(self, ctx:ModelicaParser.If_statementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#for_equation.
    def enterFor_equation(self, ctx:ModelicaParser.For_equationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#for_equation.
    def exitFor_equation(self, ctx:ModelicaParser.For_equationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#for_statement.
    def enterFor_statement(self, ctx:ModelicaParser.For_statementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#for_statement.
    def exitFor_statement(self, ctx:ModelicaParser.For_statementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#for_indices.
    def enterFor_indices(self, ctx:ModelicaParser.For_indicesContext):
        pass

    # Exit a parse tree produced by ModelicaParser#for_indices.
    def exitFor_indices(self, ctx:ModelicaParser.For_indicesContext):
        pass


    # Enter a parse tree produced by ModelicaParser#for_index.
    def enterFor_index(self, ctx:ModelicaParser.For_indexContext):
        pass

    # Exit a parse tree produced by ModelicaParser#for_index.
    def exitFor_index(self, ctx:ModelicaParser.For_indexContext):
        pass


    # Enter a parse tree produced by ModelicaParser#while_statement.
    def enterWhile_statement(self, ctx:ModelicaParser.While_statementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#while_statement.
    def exitWhile_statement(self, ctx:ModelicaParser.While_statementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#when_equation.
    def enterWhen_equation(self, ctx:ModelicaParser.When_equationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#when_equation.
    def exitWhen_equation(self, ctx:ModelicaParser.When_equationContext):
        pass


    # Enter a parse tree produced by ModelicaParser#when_statement.
    def enterWhen_statement(self, ctx:ModelicaParser.When_statementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#when_statement.
    def exitWhen_statement(self, ctx:ModelicaParser.When_statementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#connect_clause.
    def enterConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#connect_clause.
    def exitConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expression_simple.
    def enterExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expression_simple.
    def exitExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expression_if.
    def enterExpression_if(self, ctx:ModelicaParser.Expression_ifContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expression_if.
    def exitExpression_if(self, ctx:ModelicaParser.Expression_ifContext):
        pass


    # Enter a parse tree produced by ModelicaParser#simple_expression.
    def enterSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#simple_expression.
    def exitSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_neg.
    def enterExpr_neg(self, ctx:ModelicaParser.Expr_negContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_neg.
    def exitExpr_neg(self, ctx:ModelicaParser.Expr_negContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_add.
    def enterExpr_add(self, ctx:ModelicaParser.Expr_addContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_add.
    def exitExpr_add(self, ctx:ModelicaParser.Expr_addContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_exp.
    def enterExpr_exp(self, ctx:ModelicaParser.Expr_expContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_exp.
    def exitExpr_exp(self, ctx:ModelicaParser.Expr_expContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_or.
    def enterExpr_or(self, ctx:ModelicaParser.Expr_orContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_or.
    def exitExpr_or(self, ctx:ModelicaParser.Expr_orContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_primary.
    def enterExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_primary.
    def exitExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_and.
    def enterExpr_and(self, ctx:ModelicaParser.Expr_andContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_and.
    def exitExpr_and(self, ctx:ModelicaParser.Expr_andContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_rel.
    def enterExpr_rel(self, ctx:ModelicaParser.Expr_relContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_rel.
    def exitExpr_rel(self, ctx:ModelicaParser.Expr_relContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_not.
    def enterExpr_not(self, ctx:ModelicaParser.Expr_notContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_not.
    def exitExpr_not(self, ctx:ModelicaParser.Expr_notContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expr_mul.
    def enterExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expr_mul.
    def exitExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_unsigned_number.
    def enterPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_unsigned_number.
    def exitPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_string.
    def enterPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_string.
    def exitPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_false.
    def enterPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_false.
    def exitPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_true.
    def enterPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_true.
    def exitPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_function.
    def enterPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_function.
    def exitPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_derivative.
    def enterPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_derivative.
    def exitPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_initial.
    def enterPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_initial.
    def exitPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_component_reference.
    def enterPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_component_reference.
    def exitPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_output_expression_list.
    def enterPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_output_expression_list.
    def exitPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_expression_list.
    def enterPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_expression_list.
    def exitPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_function_arguments.
    def enterPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_function_arguments.
    def exitPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#primary_end.
    def enterPrimary_end(self, ctx:ModelicaParser.Primary_endContext):
        pass

    # Exit a parse tree produced by ModelicaParser#primary_end.
    def exitPrimary_end(self, ctx:ModelicaParser.Primary_endContext):
        pass


    # Enter a parse tree produced by ModelicaParser#name.
    def enterName(self, ctx:ModelicaParser.NameContext):
        pass

    # Exit a parse tree produced by ModelicaParser#name.
    def exitName(self, ctx:ModelicaParser.NameContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_reference_element.
    def enterComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):
        pass

    # Exit a parse tree produced by ModelicaParser#component_reference_element.
    def exitComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):
        pass


    # Enter a parse tree produced by ModelicaParser#component_reference.
    def enterComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):
        pass

    # Exit a parse tree produced by ModelicaParser#component_reference.
    def exitComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):
        pass


    # Enter a parse tree produced by ModelicaParser#function_call_args.
    def enterFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#function_call_args.
    def exitFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#function_arguments.
    def enterFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#function_arguments.
    def exitFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#named_arguments.
    def enterNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#named_arguments.
    def exitNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#named_argument.
    def enterNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):
        pass

    # Exit a parse tree produced by ModelicaParser#named_argument.
    def exitNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):
        pass


    # Enter a parse tree produced by ModelicaParser#argument_function.
    def enterArgument_function(self, ctx:ModelicaParser.Argument_functionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#argument_function.
    def exitArgument_function(self, ctx:ModelicaParser.Argument_functionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#argument_expression.
    def enterArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):
        pass

    # Exit a parse tree produced by ModelicaParser#argument_expression.
    def exitArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):
        pass


    # Enter a parse tree produced by ModelicaParser#output_expression_list.
    def enterOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#output_expression_list.
    def exitOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#expression_list.
    def enterExpression_list(self, ctx:ModelicaParser.Expression_listContext):
        pass

    # Exit a parse tree produced by ModelicaParser#expression_list.
    def exitExpression_list(self, ctx:ModelicaParser.Expression_listContext):
        pass


    # Enter a parse tree produced by ModelicaParser#array_subscripts.
    def enterArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):
        pass

    # Exit a parse tree produced by ModelicaParser#array_subscripts.
    def exitArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):
        pass


    # Enter a parse tree produced by ModelicaParser#subscript.
    def enterSubscript(self, ctx:ModelicaParser.SubscriptContext):
        pass

    # Exit a parse tree produced by ModelicaParser#subscript.
    def exitSubscript(self, ctx:ModelicaParser.SubscriptContext):
        pass


    # Enter a parse tree produced by ModelicaParser#comment.
    def enterComment(self, ctx:ModelicaParser.CommentContext):
        pass

    # Exit a parse tree produced by ModelicaParser#comment.
    def exitComment(self, ctx:ModelicaParser.CommentContext):
        pass


    # Enter a parse tree produced by ModelicaParser#string_comment.
    def enterString_comment(self, ctx:ModelicaParser.String_commentContext):
        pass

    # Exit a parse tree produced by ModelicaParser#string_comment.
    def exitString_comment(self, ctx:ModelicaParser.String_commentContext):
        pass


    # Enter a parse tree produced by ModelicaParser#annotation.
    def enterAnnotation(self, ctx:ModelicaParser.AnnotationContext):
        pass

    # Exit a parse tree produced by ModelicaParser#annotation.
    def exitAnnotation(self, ctx:ModelicaParser.AnnotationContext):
        pass


